from .base import CommonDBConfig
from typing import Optional


class MysqlConfig(CommonDBConfig):
    """Mysql 配置"""
    host: Optional[str] = '127.0.0.1'
    port: Optional[int] = 3306
    user: Optional[str] = 'root'
    driver: Optional[str] = 'mysql+pymysql'


class AsyncMysqlConfig(CommonDBConfig):
    """Mysql async配置"""
    host: Optional[str] = '127.0.0.1'
    port: Optional[int] = 3306
    user: Optional[str] = 'root'
    driver: Optional[str] = 'mysql+aiomysql'
    is_async: Optional[bool] = True



from .base import CommonDBConfig
from typing import Optional


class RedisConfig(CommonDBConfig):
    """Redis 配置"""
    host: Optional[str] = '127.0.0.1'
    port: Optional[int] = 6379
    user: Optional[str] = ''
    driver: Optional[str] = 'redis'



import json
from typing import Union, List

from .base import DataFileLoader, DataFileExtractor


class JsonDataFileLoader(DataFileLoader):

    def _convert_config_data_to_python(self,
                                       config_data: Union[str, bytes],
                                       **kwargs) -> Union[dict, tuple, list]:
        return json.loads(config_data)


class JsonDataFileExtractor(DataFileExtractor):

    def _convert_data_to_str(self, data: Union[dict, List[dict]],
                             **kwargs) -> str:
        return json.dumps(data, ensure_ascii=False)



import random
import time
from typing import Union, Callable, Dict, List, Literal, Optional, Any

import requests
import undetected_chromedriver as uc
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver import ActionChains
from urllib.parse import urlparse


class Chrome(uc.Chrome):
    """自定义拓展uc.Chrome, 提供更便捷的方法"""

    def __init__(
        self,
        options=None,
        user_data_dir=None,
        driver_executable_path=None,
        browser_executable_path=None,
        port=0,
        enable_cdp_events=False,
        desired_capabilities=None,
        advanced_elements=False,
        keep_alive=True,
        log_level=0,
        headless=True,
        version_main=None,
        patcher_force_close=False,
        suppress_welcome=True,
        use_subprocess=True,
        debug=False,
        no_sandbox=True,
        user_multi_procs: bool = False,
        **kw,
    ):
        # 内置常用options
        if not options:
            options = uc.ChromeOptions()
            options.add_argument("--disable-extensions")
            options.add_argument('--disable-application-cache')
            options.add_argument('--disable-gpu')
            options.add_argument("--no-sandbox")
            options.add_argument("--disable-setuid-sandbox")
            options.add_argument("--disable-dev-shm-usage")
        super().__init__(options=options,
                         user_data_dir=user_data_dir,
                         driver_executable_path=driver_executable_path,
                         browser_executable_path=browser_executable_path,
                         port=port,
                         enable_cdp_events=enable_cdp_events,
                         desired_capabilities=desired_capabilities,
                         advanced_elements=advanced_elements,
                         keep_alive=keep_alive,
                         log_level=log_level,
                         headless=headless,
                         version_main=version_main,
                         patcher_force_close=patcher_force_close,
                         suppress_welcome=suppress_welcome,
                         use_subprocess=use_subprocess,
                         debug=debug,
                         no_sandbox=no_sandbox,
                         user_multi_procs=user_multi_procs,
                         **kw)

    def wait_el(self, locator: str, by: str = uc.By.XPATH, timeout: int = 10) -> uc.WebElement:
        """等待timeout秒并获取1个元素, 获取失败则产生timeout异常"""
        WebDriverWait(self, timeout).until(
            EC.presence_of_element_located((by, locator))
        )
        return self.find_element(by, locator)

    def wait_els(self, locator: str, by: str = uc.By.XPATH, timeout: int = 10) -> List[uc.WebElement]:
        """等待timeout秒并获取符合条件的元素列表, 获取失败则产生timeout异常"""
        WebDriverWait(self, timeout).until(
            EC.presence_of_element_located((by, locator))
        )
        return self.find_elements(by, locator)

    @classmethod
    def send_keys(cls, el: uc.WebElement, content: str, interval: Union[float, Callable[[], float]] = random.random):
        """给元素el输入内容, 每输入一个字符等待interval的间隔, 更拟人化

        Args:
            el: 输入的元素
            content: 输入的内容
            interval: 输入每个元素等待的间隔
        """
        if not content:
            return
        for c in content:
            el.send_keys(c)
            if isinstance(interval, float):
                time.sleep(interval)
            else:
                time.sleep(interval())

    def action_chains(self):
        """获取当前行为链完成各种行为"""
        return ActionChains(self)

    def get_cookies_map(self) -> Dict[str, str]:
        """获取cookie key: value映射的字典"""
        ck_dict = {}
        for item in self.get_cookies():
            ck_dict[item['name']] = item['value']
        return ck_dict

    def get_cookies_str(self) -> str:
        """获取拼接字符串的cookie, 示例: x=1; y=2"""
        return '; '.join([f'{k}={v}' for k, v in self.get_cookies_map().items()])

    def invoke(self,
               url,
               method: Literal['GET', 'POST', 'PUT', 'DELETE', 'HEAD', 'OPTIONS', 'PATCH', 'CONNECT', 'TRACE'],
               params: Dict[str, Any] = None,
               data: Any = None,
               json: Dict[str, Any] = None,
               files: Any = None,
               auth: Any = None,
               timeout=None,
               proxies=None,
               stream=None,
               verify=None,
               cert=None,
               headers: Optional[Dict[str, str]] = None):
        """以http方式直接调用特定接口, 自动填充headers中不存在的ua, cookie等信息

        Args:
            url: 调用的链接
            method: 调用的方法
            params: url参数
            data: 传递的数据
            json: Content-Type 为 application/json 传递的数据
            files: 传递的文件数据
            auth: 认证信息
            headers: 覆盖的请求头, 不存在的键将使用当前环境覆盖, 除非特别声明值为None
            timeout: 超时时间, 默认不超时
            proxies: 代理对象
            stream: 是否以流方式处理响应
            verify: 是否TLS校验
            cert: TLS cert证书
        """
        headers = {k.lower(): v for k, v in headers.items()} if headers else {}
        if 'user-agent' not in headers:
            headers['user-agent'] = self.execute_script('return navigator.userAgent')
        if 'cookie' not in headers:
            headers['cookie'] = self.get_cookies_str()
        if 'sec-ch-ua' not in headers:
            brands = self.execute_script('return navigator.userAgentData.brands')
            if brands:
                sec_ch_uas = []
                for brand in brands:
                    if 'brand' not in brand or 'version' not in brand:
                        continue
                    sec_ch_uas.append(f'"{brand["brand"]}";v="{brand["version"]}"')
                if sec_ch_uas:
                    headers['sec-ch-ua'] = ', '.join(sec_ch_uas)
        if 'sec-ch-ua-mobile' not in headers:
            is_mobile = self.execute_script('return navigator.userAgentData.mobile')
            headers['sec-ch-ua-mobile'] = f'?{"1" if is_mobile else "0"}'
        if 'sec-ch-ua-platform' not in headers:
            platform = self.execute_script('return navigator.userAgentData.platform')
            if platform:
                headers['sec-ch-ua-platform'] = f'"{platform}"'
        if 'referer' not in headers:
            headers['referer'] = self.current_url
        if 'origin' not in headers:
            parsed_url = urlparse(url)
            headers['origin'] = f'{parsed_url.scheme}://{parsed_url.hostname}'
        return requests.request(method, url, params=params, data=data, json=json, headers=headers, files=files,
                                auth=auth, stream=stream, timeout=timeout, proxies=proxies, verify=verify, cert=cert)



from typing import Optional

from seatools.logger import get_loguru_adapter_logging_formatter


def setup_uvicorn(file_name,
                  rotation_type: str = 'd',
                  rotation: int = 1,
                  serialize: bool = True,
                  retention_count: int = 3,
                  level: str = 'INFO',
                  extra: Optional[dict] = None):
    """设置uvicorn文件记录, 与loguru相同的日志格式.

    Args:
        file_name: 日志文件路径及名称
        rotation:
        rotation_type:
        serialize: 是否序列化, 仅为True时生效
        retention_count:
        level: 日志级别
        extra: 额外信息
    """
    from uvicorn.config import LOGGING_CONFIG
    # 修改默认日志格式
    LOGGING_CONFIG['formatters']['default'][
        'fmt'] = '%(asctime)s | %(levelname)-8s | %(module)s.%(funcName)s:%(lineno)d - %(message)s'
    LOGGING_CONFIG['formatters']['access'][
        'fmt'] = '%(asctime)s | %(levelname)-8s | %(module)s.%(funcName)s:%(lineno)d - %(client_addr)s - “%(request_line)s” - %(status_code)s'

    # 开启序列化则注入
    if serialize:
        # 增加一个适配loguru序列化的日志格式化器
        file_formatter = {
            "()": get_loguru_adapter_logging_formatter(),
            'fmt': '%(message)s',
            'extra': extra or {},
        }
        # 增加一个文件handler
        file_handler = {
            "class": "logging.handlers.TimedRotatingFileHandler",
            "formatter": "file",
            "filename": file_name,
            "when": rotation_type,
            "interval": rotation,
            "backupCount": retention_count,
            'level': level,
            'encoding': 'utf-8'
        }
        # 注入
        LOGGING_CONFIG['formatters']['file'] = file_formatter
        LOGGING_CONFIG['handlers']['file'] = file_handler
        LOGGING_CONFIG['loggers']['uvicorn']['handlers'].append('file')
        LOGGING_CONFIG['loggers']['uvicorn.access']['handlers'].append('file')



from .base import CommonDBConfig
from typing import Optional


class PostgresqlConfig(CommonDBConfig):
    """postgresql 配置"""
    host: Optional[str] = '127.0.0.1'
    port: Optional[int] = 5432
    user: Optional[str] = 'root'
    driver: Optional[str] = 'postgresql+psycopg2'


class AsyncPostgresqlConfig(CommonDBConfig):
    """postgresql async配置"""
    host: Optional[str] = '127.0.0.1'
    port: Optional[int] = 5432
    user: Optional[str] = 'root'
    driver: Optional[str] = 'postgresql+asyncpg'
    is_async: Optional[bool] = True



from seatools import exceptions
from loguru import logger


def ex(d: int):
    try:
        logger.info(1 / d)
    except Exception as e:
        logger.error('异常')
        raise e


def test_uncaught_exception():
    exceptions.do_uncaught_exceptions(ex, 1, uncaught_exceptions=ImportError)
    exceptions.do_uncaught_exceptions(ex, 0, uncaught_exceptions=ZeroDivisionError)





def deep_update(d1: dict, d2: dict) -> dict:
    """深度合并两个配置, 仅覆盖最小层级, 重复的key用cfg2覆盖cfg1.

    注意: 该方法会覆盖d1元数据, 不会新生成字典
    """
    if not d2:
        return d1
    for key, value in d2.items():
        if key in d1 and isinstance(d1[key], dict) and isinstance(value, dict):
            deep_update(d1[key], value)
        else:
            d1[key] = value
    return d1




from typing import TypeVar, Optional, List, Union, Any, Type, Sequence, Literal, Dict

from sqlalchemy import select, ColumnElement, update, Select, Result, func, BinaryExpression, delete
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session
from sqlalchemy.engine import ScalarResult
from sqlalchemy.sql.elements import SQLCoreOperations

_BASE = TypeVar('_BASE', bound='Base')
_SQLSelectOperations = Union[Type[_BASE], SQLCoreOperations[_BASE], BinaryExpression, List[BinaryExpression], List[SQLCoreOperations[_BASE]]]
_SQLCondOperations = Union[SQLCoreOperations[_BASE], bool]  # !注意, 不能传递bool类型, 仅为了兼容编译器解析问题
_SQLOrderOperations = Union[SQLCoreOperations[_BASE], ColumnElement[_BASE]]
_SQLGroupOperations = Union[SQLCoreOperations[_BASE], ColumnElement[_BASE]]
_SQLHavingOperations = Union[SQLCoreOperations[_BASE], ColumnElement[_BASE]]


def _wrapper_to_list(item: Union[List[Any], Any]):
    if isinstance(item, list):
        return item
    return [item]


def _gen_model_query_select(entities: _SQLSelectOperations,
                            conds: Optional[Union[List[_SQLCondOperations], _SQLCondOperations]] = None,
                            orders: Optional[Union[List[_SQLOrderOperations], _SQLOrderOperations]] = None,
                            groups: Optional[Union[List[_SQLGroupOperations], _SQLGroupOperations]] = None,
                            group_conds: Optional[Union[List[_SQLHavingOperations], _SQLHavingOperations]] = None,
                            offset: Optional[int] = None,
                            limit: Optional[int] = None,
                            ) -> Select[_BASE]:
    q = select(*(_wrapper_to_list(entities)))
    if conds is not None:
        q = q.where(*(_wrapper_to_list(conds)))
    if orders is not None:
        q = q.order_by(*(_wrapper_to_list(orders)))
    if groups is not None:
        q = q.group_by(*(_wrapper_to_list(groups)))
    if group_conds is not None:
        q = q.having(*(_wrapper_to_list(group_conds)))
    if offset is not None:
        q = q.offset(offset)
    if limit is not None:
        q = q.limit(limit)
    return q


def _get_result(result: Result[_BASE],
                result_analysis: Literal[
                    'scalar', 'scalars', 'list', 'scalars-list', 'first', 'scalars-first', 'one_or_none', 'scalars-one_or_none', 'one', 'scalars-one', None] = 'scalars-list'):
    ans = result
    result_steps = result_analysis.split('-')
    for result_step in result_steps:
        if result_step == 'scalars':
            ans = ans.scalars()
        if result_step == 'list':
            ans = ans.all()
        if result_step == 'first':
            ans = ans.first()
        if result_step == 'scalar':
            ans = ans.scalar()
        if result_step == 'one_or_none':
            ans = ans.one_or_none()
        if result_step == 'one':
            ans = ans.one()
    return ans


class BaseMapper:
    """提供表的基础操作封装

    注意, 如果__tableclass__存在字段名称name与属性名称不一致, 则必须设置字段key参数且值内容与属性名称一致, 否则可能出现未知错误
    """

    # 表类型, 需要继承 DeclarativeBase
    __tableclass__: Type[_BASE] = None
    # 是否自动提交, 为True将会自动提交
    __autocommit__: bool = False

    def __init__(self):
        self._table = self.__tableclass__.__table__
        self._columns = []
        self._attr_column_map = {}
        self._primary_keys = []
        self._exclude_primary_columns = []
        for column in self._table.columns:
            self._columns.append(column.key)
            self._attr_column_map[column.key] = column.name
            if column.primary_key:
                self._primary_keys.append(column.key)
            else:
                self._exclude_primary_columns.append(column.key)

    def save(self, session: Session, instance: _BASE, ignore_none: bool = True) -> bool:
        """新增或修改

        Args:
            session: 数据库session
            instance: 实例对象
            ignore_none: 忽略None值, 为True时新增和删除将会忽略none字段
        """
        _type: Literal['add', 'update', None] = None
        # 按照所有主键查询是否存在
        conds = []
        for primary_key in self._primary_keys:
            conds.append(getattr(self.__tableclass__, primary_key) == getattr(instance, primary_key))
        # 查询, 存在则修改, 否则新增
        item = self.query(session, conds=conds, limit=1, result='scalars-first')
        if not item:
            session.add(instance)
            self._check_commit(session)
        else:
            if not conds:
                raise ValueError('表{}不存在主键, 无法通过该方法更新数据'.format(self.__tableclass__.__tablename__))
            sets = {}
            for attr_name in self._exclude_primary_columns:
                attr = getattr(instance, attr_name)
                if attr is not None or not ignore_none:
                    sets[attr_name] = attr
            if sets:
                self.update(session, sets=sets, conds=conds)
        return True

    def query_one(self, session: Session,
                  conds: Optional[Union[List[_SQLCondOperations], _SQLCondOperations]] = None,
                  orders: Optional[Union[List[_SQLOrderOperations], _SQLOrderOperations]] = None,
                  groups: Optional[Union[List[_SQLGroupOperations], _SQLGroupOperations]] = None,
                  group_conds: Optional[Union[List[_SQLHavingOperations], _SQLHavingOperations]] = None,
                  offset: Optional[int] = None,
                  limit: Optional[int] = None,
                  fields: Optional[_SQLSelectOperations] = None,
                  result: Literal[
                      'first', 'scalars-first', 'one_or_none', 'scalars-one_or_none', 'one', 'scalars-one', None] = 'scalars-first',
                  **kwargs) -> Union[ScalarResult[_BASE], Sequence[_BASE], _BASE, None, Any]:
        """条件查询, 仅返回单条记录, query的部分分支, 具体见query方法"""
        return self.query(session, conds, orders, groups, group_conds, offset, limit, fields, result, **kwargs)

    def query(self, session: Session,
              conds: Optional[Union[List[_SQLCondOperations], _SQLCondOperations]] = None,
              orders: Optional[Union[List[_SQLOrderOperations], _SQLOrderOperations]] = None,
              groups: Optional[Union[List[_SQLGroupOperations], _SQLGroupOperations]] = None,
              group_conds: Optional[Union[List[_SQLHavingOperations], _SQLHavingOperations]] = None,
              offset: Optional[int] = None,
              limit: Optional[int] = None,
              query: Optional[_SQLSelectOperations] = None,
              result: Literal[
                  'scalar', 'scalars', 'list', 'scalars-list', 'first', 'scalars-first', 'one_or_none', 'scalars-one_or_none', 'one', 'scalars-one', None] = 'scalars-list',
              **kwargs) -> Union[ScalarResult[_BASE], Sequence[_BASE], _BASE, None, Any]:
        """条件查询

        Args:
            session: 数据库session
            conds: 条件, 例如 XXXTable.id == 2, XXXTable.id.in_(1, 2, 3), 传递该形式的条件
            orders: 排序, 例如 XXXTable.id.asc(), XXXTable.id.desc() 传递该形式参数
            groups: 分组
            group_conds: 分组条件, 对应 sql having
            offset: offset
            limit: limit
            query: 查询内容, 为 None 默认查询表所有字段, 允许传数据表属性字段及列表, 允许传递sqlalchemy.func函数及函数列表
            result: 返回数据形式, 默认 scalars_list, 各类型见源码

        Returns:
            ScalarResult或者List
        """
        q = _gen_model_query_select(query if query is not None else self.__tableclass__,
                                    conds=conds,
                                    orders=orders,
                                    groups=groups,
                                    group_conds=group_conds,
                                    offset=offset,
                                    limit=limit)
        ans = session.execute(q)
        result = _get_result(ans, result_analysis=result)
        self._check_commit(session)
        return result

    def update(self, session: Session,
               sets: Dict[str, Any],
               conds: Optional[Union[List[_SQLCondOperations], _SQLCondOperations]] = None):
        """修改

        Args:
            session: 数据库session
            sets: 需要修改的字段值字典, 例如修改名称为xxx1, {'name': 'xxx1'}
            conds: 需要修改的属性的查询条件
        """
        up = update(self.__tableclass__).values(**sets)
        if conds:
            conds = _wrapper_to_list(conds)
            up = up.where(*conds)
        session.execute(up)
        self._check_commit(session)
        return True

    def delete(self, session: Session,
               conds: Optional[Union[List[_SQLCondOperations], _SQLCondOperations]] = None):
        """删除

        Args:
            session: 数据库session
            conds: 查询条件
        """
        if not conds:
            raise ValueError('删除数据查询条件不能为空')
        dt = delete(self.__tableclass__).where(*(_wrapper_to_list(conds)))
        session.execute(dt)
        self._check_commit(session)
        return True

    def _check_commit(self, session: Session):
        if self.__autocommit__:
            session.commit()


class AsyncBaseMapper:
    """提供表的基础操作封装"""

    # 表类型, 需要继承 DeclarativeBase
    __tableclass__: Type[_BASE] = None
    # 是否自动提交, 为True将会自动提交
    __autocommit__: bool = False

    def __init__(self):
        self._table = self.__tableclass__.__table__
        self._columns = []
        self._attr_column_map = {}
        self._primary_keys = []
        self._exclude_primary_columns = []
        for column in self._table.columns:
            self._columns.append(column.key)
            self._attr_column_map[column.key] = column.name
            if column.primary_key:
                self._primary_keys.append(column.key)
            else:
                self._exclude_primary_columns.append(column.key)

    async def save(self, session: AsyncSession, instance: _BASE, ignore_none: bool = True) -> bool:
        """新增或修改

        Args:
            session: 数据库session
            instance: 实例对象
            ignore_none: 忽略None值, 为True时新增和删除将会忽略none值
        """
        _type: Literal['add', 'update', None] = None
        # 按照所有主键查询是否存在
        conds = []
        for primary_key in self._primary_keys:
            conds.append(getattr(self.__tableclass__, primary_key) == getattr(instance, primary_key))
        # 查询, 存在则修改, 否则新增
        item = await self.query(session, conds=conds, limit=1, result='scalars-first')
        if not item:
            session.add(instance)
            await self._check_commit(session)
        else:
            if not conds:
                raise ValueError('表{}不存在主键, 无法通过该方法更新数据'.format(self.__tableclass__.__tablename__))
            sets = {}
            for attr_name in self._exclude_primary_columns:
                attr = getattr(instance, attr_name)
                if attr is not None or not ignore_none:
                    sets[self._attr_column_map[attr_name]] = attr
            if sets:
                await self.update(session, sets=sets, conds=conds)
        return True

    async def query_one(self, session: AsyncSession,
                        conds: Optional[Union[List[_SQLCondOperations], _SQLCondOperations]] = None,
                        orders: Optional[Union[List[_SQLOrderOperations], _SQLOrderOperations]] = None,
                        groups: Optional[Union[List[_SQLGroupOperations], _SQLGroupOperations]] = None,
                        group_conds: Optional[Union[List[_SQLHavingOperations], _SQLHavingOperations]] = None,
                        offset: Optional[int] = None,
                        limit: Optional[int] = None,
                        fields: Optional[_SQLSelectOperations] = None,
                        result: Literal[
                            'first', 'scalars-first', 'one_or_none', 'scalars-one_or_none', 'one', 'scalars-one', None] = 'scalars-first',
                        **kwargs) -> Union[ScalarResult[_BASE], Sequence[_BASE], _BASE, None, Any]:
        """条件查询, 仅返回单条记录, query的部分分支, 具体见query方法"""
        return await self.query(session, conds, orders, groups, group_conds, offset, limit, fields, result, **kwargs)

    async def query(self, session: AsyncSession,
                    conds: Optional[Union[List[_SQLCondOperations], _SQLCondOperations]] = None,
                    orders: Optional[Union[List[_SQLOrderOperations], _SQLOrderOperations]] = None,
                    groups: Optional[Union[List[_SQLGroupOperations], _SQLGroupOperations]] = None,
                    group_conds: Optional[Union[List[_SQLHavingOperations], _SQLHavingOperations]] = None,
                    offset: Optional[int] = None,
                    limit: Optional[int] = None,
                    query: Optional[_SQLSelectOperations] = None,
                    result: Literal[
                        'scalar', 'scalars', 'list', 'scalars-list', 'first', 'scalars-first', 'one_or_none', 'scalars-one_or_none', 'one', 'scalars-one', None] = 'scalars-list',
                    **kwargs) -> Union[ScalarResult[_BASE], Sequence[_BASE], _BASE, None, Any]:
        """条件查询

        Args:
            session: 数据库session
            conds: 条件, 例如 XXXTable.id == 2, XXXTable.id.in_(1, 2, 3), 传递该形式的条件
            orders: 排序, 例如 XXXTable.id.asc(), XXXTable.id.desc() 传递该形式参数
            groups: 分组
            group_conds: 分组条件
            offset: offset
            limit: limit
            query: 查询内容, 为 None 默认查询表所有字段, 允许传数据表属性字段及列表, 允许传递sqlalchemy.func函数及函数列表
            result: 返回数据形式, 默认 scalars_list, 各类型见源码

        Returns:
            ScalarResult或者List
        """
        q = _gen_model_query_select(query if query is not None else self.__tableclass__,
                                    conds=conds,
                                    orders=orders,
                                    groups=groups,
                                    group_conds=group_conds,
                                    offset=offset,
                                    limit=limit)
        ans = await session.execute(q)
        result = _get_result(ans, result_analysis=result)
        await self._check_commit(session)
        return result

    async def update(self, session: AsyncSession,
                     sets: Dict[str, Any],
                     conds: Optional[Union[List[_SQLCondOperations], _SQLCondOperations]] = None):
        """修改

        Args:
            session: 数据库session
            sets: 需要修改的字段值字典, 例如修改名称为xxx1, {'name': 'xxx1'}
            conds: 需要修改的属性的查询条件
        """
        up = update(self.__tableclass__).values(**sets)
        if conds:
            conds = _wrapper_to_list(conds)
            up = up.where(*conds)
        await session.execute(up)
        await self._check_commit(session)
        return True

    async def delete(self, session: AsyncSession,
               conds: Optional[Union[List[_SQLCondOperations], _SQLCondOperations]] = None):
        """删除

        Args:
            session: 数据库session
            conds: 查询条件
        """
        if not conds:
            raise ValueError('删除数据查询条件不能为空')
        dt = delete(self.__tableclass__).where(*(_wrapper_to_list(conds)))
        await session.execute(dt)
        await self._check_commit(session)
        return True

    async def _check_commit(self, session: AsyncSession):
        if self.__autocommit__:
            await session.commit()



from .base import CommonDBConfig
from typing import Optional


class ClickhouseConfig(CommonDBConfig):
    """Clickhouse 配置"""
    host: Optional[str] = '127.0.0.1'
    port: Optional[int] = 8123
    user: Optional[str] = 'root'
    driver: Optional[str] = 'clickhouse'



from cachetools import Cache as _Cache
from seatools.sqlalchemy import SqlAlchemyClient
from seatools.sqlalchemy import Base
from sqlalchemy import String
from sqlalchemy.dialects.sqlite import insert
from sqlalchemy.orm import mapped_column


class _CacheData(Base):
    __tablename__ = 'cache'

    key = mapped_column(String, primary_key=True)
    value = mapped_column(String)


class SqliteCache(_Cache):
    """Sqlalchemy的sqlite缓存实现"""

    def __init__(self, maxsize, conn_url='sqlite:///cache.db', getsizeof=None):
        super().__init__(maxsize, getsizeof)
        self._conn_url = conn_url

    def _sqlite_cli(self) -> SqlAlchemyClient:
        if hasattr(self, '_sqlite_client') and self._sqlite_client:
            return self._sqlite_client
        self._sqlite_client = SqlAlchemyClient(self._conn_url)
        Base.metadata.create_all(self._sqlite_client.engine())
        return self._sqlite_client

    def __contains__(self, key):
        with self._sqlite_cli().session() as session:
            return session.query(_CacheData).where(_CacheData.key == key).count() > 0

    def __getitem__(self, key):
        with self._sqlite_cli().session() as session:
            cache_data = session.query(_CacheData).where(_CacheData.key == key).first()
            if cache_data:
                return cache_data.value
            raise KeyError(key)

    def __setitem__(self, key, value):
        with self._sqlite_cli().session() as session:
            session.execute(insert(_CacheData).values(key=key, value=value).on_conflict_do_update(
                index_elements=['key'],
                set_={'value': value}
            ))

    def __delitem__(self, key):
        with self._sqlite_cli().session() as session:
            session.query(_CacheData).where(_CacheData.key == key).delete()



from typing import Optional

from .logging_setup import setup_logging


def setup_sqlalchemy(file_name,
                     rotation_type: str = 'd',
                     rotation: int = 1,
                     serialize: bool = True,
                     retention_count: int = 3,
                     level: str = 'INFO',
                     extra: Optional[dict] = None):
    """设置sqlalchemy日志记录, 与loguru相同的日志格式."""
    setup_logging(file_name=file_name,
                  logger_name='sqlalchemy',
                  rotation_type=rotation_type,
                  rotation=rotation,
                  serialize=serialize,
                  retention_count=retention_count,
                  level=level,
                  extra=extra)



import abc
from typing import Any


class BasicTypeMixin(abc.ABC):
    """基本数据类型Mixin."""
    @property
    def value(self):
        raise NotImplementedError()

    @property
    def int(self):
        return int(self.value)

    @property
    def float(self):
        return float(self.value)

    @property
    def bool(self):
        return bool(self.value)

    @property
    def str(self):
        return str(self.value)

    @property
    def list(self):
        return list(self.value)

    @property
    def dict(self):
        return dict(self.value)

    @property
    def tuple(self):
        return tuple(self.value)

    @property
    def set(self):
        return set(self.value)

    @property
    def complex(self):
        return complex(self.value)


class BaseBeanProxy(BasicTypeMixin, abc.ABC):
    """基础代理, 未提供代理逻辑实现, 仅提供抽象封装"""
    _exclude_attr = []

    def __init__(self, name: str, obj: Any, *args, **kwargs):
        """

        Args:
             name: 代理bean名称
             obj: 需要被代理的对象
        """
        self._name = name
        self._obj = obj
        self._extra_args = args
        self._extra_kwargs = kwargs

    def ioc_name(self) -> str:
        """获取bean名称"""
        return self._name

    def ioc_type(self):
        """获取被代理的类型"""
        return type(self.ioc_bean())

    def ioc_bean(self):
        """获取被代理的真实对象"""
        return self._obj

    def __getattr__(self, item):
        return getattr(self.ioc_bean(), item)

    def __getattribute__(self, item):
        if item.startswith('__') and item.endswith('__'):
            attr = getattr(self.ioc_bean(), item)
            if callable(attr):
                def wrapper(*args, **kwargs):
                    return attr(*args, **kwargs)

                return wrapper
            return attr
        return object.__getattribute__(self, item)

    def __call__(self, *args, **kwargs):
        return self.ioc_bean()(*args, **kwargs)

    def __repr__(self):
        return self.ioc_bean().__repr__()

    @property
    def value(self):
        return self.ioc_bean()


class ClassBeanProxy(BaseBeanProxy):
    """类对象bean代理"""

    def __init__(self, name: str, obj: Any, *args, primary: bool = False, **kwargs):
        """
        Args:
            obj: 特定类的对象
        """
        super().__init__(name, obj, *args, **kwargs)
        self._primary = primary

    def instanceof(self, clazz):
        """判断实际对象是否是某个类型的实例"""
        return isinstance(self.ioc_bean(), clazz)

    def primary(self):
        return self._primary



import time
import threading


class Snowflake:
    """雪花ID生成工具

    使用方式:
        s = Snowflake(datacenter_id=1, worker_id=1) # 根据业务场景给不同机器设置不同的datacenter_id与worker_id
        unique_id = s.next_id() # 获取ID
    """

    def __init__(self, datacenter_id: int = 1, worker_id: int = 1):
        """

        Args:
            datacenter_id: 数据中心ID
            worker_id: 工单节点ID
        """
        self._datacenter_id = datacenter_id
        self._worker_id = worker_id
        self._epoch = 1288834974657
        self._timestamp_left_shift = 22
        self._sequence = 0
        self._last_timestamp = -1
        # 初始化锁
        self._lock = threading.Lock()

    def next_id(self):
        with self._lock:  # 使用锁来同步线程
            current_timestamp = self._get_timestamp()
            if current_timestamp < self._last_timestamp:
                raise Exception("Clock moved backwards. Refusing to generate id")

            if current_timestamp == self._last_timestamp:
                self._sequence = (self._sequence + 1) & 0xfff
                if self._sequence == 0:
                    current_timestamp = self._til_next_millis(self._last_timestamp)
            else:
                self._sequence = 0

            self._last_timestamp = current_timestamp

            return ((current_timestamp - self._epoch) << self._timestamp_left_shift) | \
                (self._datacenter_id << 12) | \
                (self._worker_id << 7) | \
                self._sequence

    def _get_timestamp(self):
        return int(time.time() * 1000)

    def _til_next_millis(self, _last_timestamp):
        while True:
            current_timestamp = self._get_timestamp()
            if current_timestamp > _last_timestamp:
                return current_timestamp



import json
from typing import Optional
from seatools.models import BaseModel
from seatools.files import AutoDataFileLoader, DataType


class TestBModel(BaseModel):
    c: Optional[int] = None
    d: Optional[str] = None


class TestModel(BaseModel):
    a: Optional[int] = None
    b: Optional[TestBModel] = None


def test_file():
    adfl = AutoDataFileLoader()
    # yaml 测试
    print(adfl.load("""
    a: 1
    b:
      c: 2
      d: sadas
    """))
    print(adfl.load("""
        a: 1
        b:
          c: 2
          d: sadas
        """, TestModel))
    print(adfl.load(str({'a': 1, 'b': {'c': 2, 'd': 'sadas'}})))
    print(adfl.load(str({'a': 1, 'b': {'c': 2, 'd': 'sadas'}}), TestModel))
    print(adfl.load(json.dumps({'a': 1, 'b': {'c': 2, 'd': 'sadas'}})))
    print(adfl.load(json.dumps(
        {'a': 1, 'b': {'c': 2, 'd': 'sadas'}}), TestModel))
    print(adfl.load("""
        [b]
        c = 2
        d = sadas
        """))
    print(adfl.load("""
    [b]
    c = 2
    d = sadas
    """, TestModel))

    print(adfl.load("""a,c,d
1,2,3
4,5,6
7,8,9""", header=True))
    print(adfl.load("""a,c,d
1,2,3
4,5,6
7,8,9""", modelclass=TestModel, header=True, data_type=DataType.CSV))
    print(adfl.load("""
b.c=123
b.d=s3434
"""))
    print(adfl.load("""
    b.c=123
    b.d=s3434
    """, modelclass=TestModel))



from seatools.env import get_env
from seatools.files import AutoDataFileLoader
from typing import Optional, Dict
from loguru import logger
import os
import types


class _Properties:
    cfg: Optional[Dict] = types.MappingProxyType({})


def load_config(config_dir: str):
    """仅加载`config_dir`中的application.yml, application-dev.yml, application-test.yml, application-pro.yml
    :param config_dir: 项目目录地址
    """
    # 当前项目的配置文件
    config_file_path = config_dir + os.sep + f'application.yml'
    # 多环境配置文件
    env_config_file_path = config_dir + os.sep + f'application-{get_env().name}.yml'
    exist_config_file_path, exist_env_config_file_path = os.path.exists(config_file_path), os.path.exists(
        env_config_file_path)
    _data_file_loader = AutoDataFileLoader()
    # 配置对象_cfg
    _cfg_dict = {}
    if exist_config_file_path:
        _cfg_dict = _merge_config(_cfg_dict, _data_file_loader.load_file(config_file_path,
                                                                         encoding='utf-8'))
        logger.info('加载[{}]配置文件', config_file_path)
    if exist_env_config_file_path:
        _cfg_dict = _merge_config(_cfg_dict, _data_file_loader.load_file(env_config_file_path,
                                                                         encoding='utf-8'))
        logger.info('加载[{}]配置文件', env_config_file_path)
    # 转为不可读配置信息
    _Properties.cfg = types.MappingProxyType(_cfg_dict)


def _merge_config(cfg1: dict, cfg2: dict) -> dict:
    """ 合并两个配置, 重复的key用cfg2覆盖cfg1 """
    if not cfg2:
        return cfg1
    for key, value in cfg2.items():
        if key in cfg1 and isinstance(cfg1[key], dict) and isinstance(value, dict):
            _merge_config(cfg1[key], value)
        else:
            cfg1[key] = value
    return cfg1


def cfg() -> Optional[Dict]:
    """
    获取配置对象, 需要先调用 load_config, 否则返回空
    Returns:
        配置对象
    """
    return _Properties.cfg



from sqlalchemy import create_engine, Engine, text
from sqlalchemy.orm import sessionmaker, DeclarativeBase, Session
from sqlalchemy.ext.asyncio import create_async_engine, async_sessionmaker, AsyncEngine, AsyncSession
from contextlib import contextmanager, asynccontextmanager
from typing import Any, Type, TypeVar
from pydantic import BaseModel

_T = TypeVar('_T', bound='BaseModel')
_BASE = TypeVar('_BASE', bound='Base')


class Base(DeclarativeBase):
    """orm model 基础类"""

    def to_dict(self, exclude_none: bool = False, exclude_zero: bool = False) -> dict:
        """将当前model转为dict对象

        Args:
            exclude_none: 忽略为none的字段
            exclude_zero: 忽略0值的字段, 包含'', 0, False等0值
        """
        ans = {}
        for col in self.__table__.columns:
            val = getattr(self, col.name)
            if val or not exclude_none and not exclude_zero:
                ans[col.name] = val
            elif exclude_none and val is not None:
                ans[col.name] = val
        return ans

    def to_model(self, modelclass: Type[BaseModel], exclude_none: bool = False, exclude_zero: bool = False, **kw) -> _T:
        """将当前model转为pydantic的model对象

        Args:
            modelclass: 具体的模型类型
            exclude_none: 忽略为none的字段
            exclude_zero: 忽略0值的字段, 包含'', 0, False等0值
            kw: 自定义映射
        """
        model_dict = self.to_dict(exclude_none=exclude_none, exclude_zero=exclude_zero)
        for k, v in model_dict.items():
            kw.setdefault(k, v)
        return modelclass(**kw)

    @classmethod
    def from_model(cls,
                   model: BaseModel,
                   ignore_absent_fields: bool = True,
                   include: Any = None,
                   exclude: Any = None,
                   by_alias: bool = False,
                   exclude_unset: bool = False,
                   exclude_defaults: bool = False,
                   exclude_none: bool = False,
                   round_trip: bool = False,
                   ) -> _BASE:
        """将pydantic的model转为当前类型model

        Args:
            model: pydantic的模型值
            ignore_absent_fields: 忽略当前model中不存在的pydantic model的属性
            include: A list of fields to include in the output.
            exclude: A list of fields to exclude from the output.
            by_alias: Whether to use the field's alias in the dictionary key if defined.
            exclude_unset: Whether to exclude fields that have not been explicitly set.
            exclude_defaults: Whether to exclude fields that are set to their default value.
            exclude_none: Whether to exclude fields that have a value of `None`.
            round_trip: If True, dumped values should be valid as input for non-idempotent types such as Json[T].
        """
        model_dict = model.model_dump(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip
        )
        if ignore_absent_fields:
            for key in list(model_dict.keys()):
                if not hasattr(cls, key):
                    del model_dict[key]
        return cls(**model_dict)

    @classmethod
    def truncate(cls, session: Session):
        """Truncate sql."""
        return session.execute(text('TRUNCATE {}'.format(cls.__tablename__)))


class SqlAlchemyClient:
    """SqlAlchemy工具客户端"""

    def __init__(self, url: str, echo: bool = False, **kwargs: Any):
        """创建sqlalchemy工具客户端

        Args:
            url: 连接串
                示例:
                    sqlite:///dbname
                    mysql+pymysql://username:password@host:port/dbname
            echo: The echo=True parameter indicates that SQL emitted by connections will be logged to standard out.
        """
        self._engine = create_engine(url, echo=echo, **kwargs)
        self._session_maker = sessionmaker(bind=self._engine, expire_on_commit=False)

    @contextmanager
    def session(self) -> Session:
        """获取session上下文对象

        使用示例:
            client = SqlAlchemyClient('xxxx')
            with client.session() as session:
                session.execute()
                session.get()
        """
        session = self._session_maker()
        try:
            yield session
            session.commit()
        except Exception as e:
            session.rollback()
            raise e
        finally:
            if session:
                session.close()

    def engine(self) -> Engine:
        """获取引擎"""
        return self._engine


class AsyncSqlAlchemyClient:
    """SqlAlchemy异步工具客户端"""

    def __init__(self, url: str, echo: bool = False, **kwargs: Any):
        """创建sqlalchemy异步工具客户端

        Args:
            url: 连接串
            示例:
                sqlite:///dbname
                mysql+pymysql://username:password@host:port/dbname
            echo: The echo=True parameter indicates that SQL emitted by connections will be logged to standard out.
        """
        self._engine = create_async_engine(url, echo=echo, **kwargs)
        self._session_maker = async_sessionmaker(bind=self._engine)

    @asynccontextmanager
    async def session(self) -> AsyncSession:
        """获取session上下文对象

        使用示例:
            client = AsyncSqlAlchemyClient('xxxx')
            with client.session() as session:
                session.execute()
                session.get()
        """
        session = self._session_maker()
        try:
            yield session
            await session.commit()
        except Exception as e:
            await session.rollback()
            raise e
        finally:
            if session:
                await session.close()

    def engine(self) -> AsyncEngine:
        """获取引擎"""
        return self._engine



from .base import CommonDBConfig
from typing import Optional


class HiveConfig(CommonDBConfig):
    """Hive 配置"""
    host: Optional[str] = '127.0.0.1'
    port: Optional[int] = 10000
    driver: Optional[str] = 'hive'



from typing import Any, Type, TypeVar, Union

from seatools.ioc.beans.proxy import DelayAutowiredClassBeanProxy, DelayConfigAutowiredClassBeanProxy
from seatools.ioc.beans.proxy.base import BasicTypeMixin
from seatools.ioc.constants import required as default_required

_T = TypeVar('_T', bound=Any)


class Autowired:
    """依赖自动注入"""

    def __new__(_, value: str = None, *, cls: Type[_T]=None, required: bool = True) -> _T:
        """注入对象

        Args:
            value: bean name
            cls: 类型
            required: 是否必须, 为true找不到注入对象会报错
        """
        return DelayAutowiredClassBeanProxy(name=value, obj=cls, required=required)


class Value:
    """配置自动注入"""

    def __new__(_, value: str, *args, cls: Type[_T]=None, default_value=default_required) -> Union[_T, BasicTypeMixin]:
        """参数注入

        Args:
            value: 配置名称, 必须使用${}包裹, 多级使用.分隔, 示例: ${xxx.xxx}
            cls: value读取数据的类型, 如果为None则返回环境变量中的值
            default_value: 默认值, 若不传则表示必须有配置值
        """
        if not value:
            raise ValueError('value配置名称不能为空')
        return DelayConfigAutowiredClassBeanProxy(name=value, obj=cls, default_value=default_value)



import logging
from logging.handlers import TimedRotatingFileHandler
from typing import Optional

from seatools.logger import get_loguru_adapter_logging_formatter


def setup_logging(file_name: str,
                  logger_name: str,
                  rotation_type: str = 'd',
                  rotation: int = 1,
                  serialize: bool = True,
                  retention_count: int = 3,
                  level: str = 'INFO',
                  extra: Optional[dict] = None):
    """设置logging日志记录, 与loguru相同的日志格式.

    Args:
        file_name: 日志文件路径及名称
        logger_name: 日志名称
        rotation:
        rotation_type:
        serialize: 是否序列化, 仅为True时生效
        retention_count:
        level: 日志级别
        extra: 额外信息
    """
    # 开启序列化则注入
    if serialize:
        # 增加一个适配loguru序列化的日志格式化器
        formatter_cls = get_loguru_adapter_logging_formatter()
        # 增加一个文件handler
        file_handler = TimedRotatingFileHandler(
            filename=file_name,
            when=rotation_type,
            interval=rotation,
            backupCount=retention_count,
            encoding='utf-8',
        )

        file_handler.setFormatter(formatter_cls('%(message)s', extra=extra))
        file_handler.addFilter(lambda e: e.levelno >= logging._nameToLevel[level])
        logging_logger = logging.getLogger(logger_name)
        logging_logger.addHandler(file_handler)



import asyncio
import functools
import uuid
from typing import Optional
import inspect
from contextvars import ContextVar

from seatools.utils import list_utils
from seatools.ioc import Autowired
from seatools.sqlalchemy import SqlAlchemyClient, AsyncSqlAlchemyClient

__db_context_id = ContextVar('sqlalchemy_context_id')
__db_session_map = {}


def auto_session(*args, db: Optional[str] = None, field_name: Optional[str] = 'session'):
    """自动注入Session/AsyncSession装饰器, 通过ContextVar保证上下文的Session/AsyncSession一致性

    使用方法:
        ```python
            @auto_session('xxx') # 若仅存在一个db 或者使用某个primary=True的db, 可简写为 @auto_session() 或 @auto_session
            [async] def handle1(*args, session: Union[Session, AsyncSession] = None, **kwargs):
                [await] session.execute(...)
                [await] handle2(...)

            @auto_session('xxx')
            [async] def handle2(*args, session: Union[Session, AsyncSession] = None, **kwargs):
                [await] session.execute(...)

            # 执行
            handle1() # asyncio.run(handle1())
        ```
        session 由装饰器管理自动注入无需传参, 传参则使用传递的session。
        上述示例 handle1 与 handle2 相同db共用同一个session, 保障连续调用的事务一致性, 全部执行完成自动执行commit, 失败则自动执行rollback

    Args:
        db: 数据库配置, 与 Autowired 使用一致填写bean名称, 当只有一个db配置或者有注入primary的db bean时可不传该参数获取session实例
        field_name: 需要注入Session/AsyncSession的字段
    """

    def init_sqlalchemy_context_id():
        try:
            __db_context_id.get()
        except LookupError:
            context_id = uuid.uuid4()
            __db_context_id.set(context_id)

    async def do_async_func(func, *args, session, **kwargs):
        context_id = __db_context_id.get()
        __db_session_map[context_id] = session
        kwargs[field_name] = session
        ans = await func(*args, **kwargs)
        if context_id in __db_session_map:
            del __db_session_map[context_id]
        return ans

    def do_func(func, *args, session, **kwargs):
        context_id = __db_context_id.get()
        __db_session_map[context_id] = session
        kwargs[field_name] = session
        ans = func(*args, **kwargs)
        if context_id in __db_session_map:
            del __db_session_map[context_id]
        return ans

    def wrapper(func):
        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            if kwargs.get(field_name):
                return await func(*args, **kwargs)
            init_sqlalchemy_context_id()
            session = __db_session_map.get(__db_context_id.get())
            if session:
                kwargs[field_name] = session
                return await func(*args, **kwargs)
            cli = Autowired(db, cls=AsyncSqlAlchemyClient, required=False)
            # 检查是否存在bean实例
            if cli.ioc_bean():
                async with cli.session() as session:
                    return await do_async_func(func, session=session, *args, **kwargs)
            cli = Autowired(db, cls=SqlAlchemyClient, required=False)
            if cli.ioc_bean():
                with cli.session() as session:
                    return await do_async_func(func, session=session, *args, **kwargs)

            # redis
            import redis
            cli = Autowired(db, cls=redis.Redis)
            return await do_async_func(func, session=cli, *args, **kwargs)

        @functools.wraps(func)
        def sync_wrapper(*args, **kwargs):
            if kwargs.get(field_name):
                return func(*args, **kwargs)
            init_sqlalchemy_context_id()
            session = __db_session_map.get(__db_context_id.get())
            if session:
                kwargs[field_name] = session
                return func(*args, **kwargs)
            cli = Autowired(db, cls=SqlAlchemyClient, required=False)
            if cli.ioc_bean():
                with cli.session() as session:
                    return do_func(func, session=session, *args, **kwargs)

            # redis
            import redis
            cli = Autowired(db, cls=redis.Redis)
            return do_func(func, session=cli, *args, **kwargs)

        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        return sync_wrapper

    if len(args) == 1 and (inspect.isclass(args[0]) or inspect.isfunction(args[0])):
        return wrapper(args[0])

    db = list_utils.get(args, 0, db)
    field_name = list_utils.get(args, 1, field_name)

    return wrapper


def new_session(*args, db: Optional[str] = None, field_name: Optional[str] = 'session'):
    """自动注入Session/AsyncSession装饰器, 仅新建一个Session/AsyncSession, 需要自己保障事务一致性

    使用方法:
        ```python
            @new_session('xxx') # 若仅有一个db或者使用primary=True的db则可简写 @new_session() 或 @new_session
            [async] def handle1(*args, session: Union[Session, AsyncSession] = None, **kwargs):
                [await] session.execute(...)
                [await] handle2(session=session)

            [async] def handle2(*args, session: Union[Session, AsyncSession] = None, **kwargs):
                [await] session.execute(...)

            # 执行
            handle1() # asyncio.run(handle1())
        ```
        session 由装饰器生成自动注入, 与auto_session相比不具备自动注入的传递性。
        执行完成自动执行commit, 失败则自动执行rollback
    Args:
        db: 数据库配置, 与 Autowired 使用一致填写bean名称, 当只有一个db配置或者有注入primary的db bean时可不传该参数获取session实例
        field_name: 需要注入Session/AsyncSession的字段
    """
    def wrapper(func):
        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            if kwargs.get(field_name):
                return await func(*args, **kwargs)
            cli = Autowired(db, cls=AsyncSqlAlchemyClient, required=False)
            if cli.ioc_bean():
                async with cli.session() as session:
                    kwargs[field_name] = session
                    return await func(*args, **kwargs)
            cli = Autowired(db, cls=SqlAlchemyClient, required=False)
            if cli.ioc_bean():
                with cli.session() as session:
                    kwargs[field_name] = session
                    return await func(*args, **kwargs)

            # redis
            import redis
            cli = Autowired(db, cls=redis.Redis)
            return await func(func, session=cli, *args, **kwargs)

        @functools.wraps(func)
        def sync_wrapper(*args, **kwargs):
            if kwargs.get(field_name):
                return func(*args, **kwargs)
            cli = Autowired(db, cls=SqlAlchemyClient, required=False)
            if cli.ioc_bean():
                with cli.session() as session:
                    kwargs[field_name] = session
                    return func(*args, **kwargs)

            # redis
            import redis
            cli = Autowired(db, cls=redis.Redis)
            return func(func, session=cli, *args, **kwargs)

        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        return sync_wrapper

    if len(args) == 1 and (inspect.isclass(args[0]) or inspect.isfunction(args[0])):
        return wrapper(args[0])

    db = list_utils.get(args, 0, db)
    field_name = list_utils.get(args, 1, field_name)

    return wrapper



from typing import Union, TypeVar, Optional
from seatools.sqlalchemy import SqlAlchemyClient, AsyncSqlAlchemyClient
from seatools.sqlalchemy.dbconfig import CommonDBConfig, SqlalchemyConfig
from sqlalchemy import URL
from loguru import logger

__db_map = {}
REDIS_TYPE = TypeVar('REDIS_TYPE', bound='redis.Redis')
_DEFAULT_SQLALCHEMY_CONFIG = SqlalchemyConfig(echo=True, pool_recycle=3600)


def new_client(_id: CommonDBConfig, config: Optional[Union[SqlalchemyConfig, dict]] = None) -> Union[
    SqlAlchemyClient, AsyncSqlAlchemyClient, REDIS_TYPE]:
    """DB客户端创建工具, 可通过通用DB配置对象创建 SqlAlchemyClient, AsyncSqlAlchemyClient, redis.Redis 实例.

    Args:
        _id: CommonDBConfig 对象
        config: sqlalchemy配置, redis暂不支持

    Returns:
        SqlAlchemyClient|AsyncSqlAlchemyClient: DB客户端, 根    据配置的is_async决定返回同步还是异步客户端
        redis.Redis: redis客户端, 根据配置生成

    """
    db_config, _id = _id, str(__gen_sqlalchemy_url(_id))
    client = __db_map.get(_id)
    if client:
        return client
    # 支持 redis 获取 redis.Redis 连接对象
    if db_config.driver == 'redis':
        return __new_redis_client(db_config, _id)
    return __new_sqlalchemy_client(db_config, _id, config or _DEFAULT_SQLALCHEMY_CONFIG)


def __gen_sqlalchemy_url(config: CommonDBConfig):
    return URL.create(config.driver,
                      host=config.host,
                      port=config.port,
                      username=config.user,
                      password=config.password,
                      database=config.db)


def __new_redis_client(config: CommonDBConfig, _id: str) -> REDIS_TYPE:
    try:
        # 优先使用redis-om
        from unique_tools.redis_om import get_redis_connection
        client = get_redis_connection(url=__gen_sqlalchemy_url(config).render_as_string(hide_password=False))
        __db_map[_id] = client
        return client
    except ImportError:
        import redis
        url = __gen_sqlalchemy_url(config).render_as_string(hide_password=False)
        client = redis.Redis.from_url(url, **{'decode_responses': True})
        __db_map[_id] = client
        return client


def __new_sqlalchemy_client(config: CommonDBConfig, _id: str, db_config: Union[SqlalchemyConfig, dict]) -> Union[
    SqlAlchemyClient, AsyncSqlAlchemyClient]:
    url = __gen_sqlalchemy_url(config)
    logger.info('初始化ID[{}]的SqlAlchemyClient, 连接串[{}]', _id, url)
    client_cls = AsyncSqlAlchemyClient if config.is_async else SqlAlchemyClient
    if isinstance(db_config, SqlalchemyConfig):
        db_config = db_config.model_dump(mode='json')
    # hive 需要额外处理
    if config.driver == 'hive':
        try:
            from pyhive import hive
        except ImportError as e:
            logger.error('未安装pyhive依赖, sqlalchemy无法配置hive')
            raise e
        client = client_cls(url=url,
                            creator=lambda: hive.Connection(
                                host=config.host, port=config.port, username=config.user, database=config.db
                            ), **db_config)
    else:
        client = client_cls(url=url, **db_config)
    __db_map[_id] = client
    return client



from typing import Optional

from loguru import logger
from seatools.logger import setup


def setup_loguru(file_name,
                 rotation="1 days",
                 serialize=True,
                 backtrace=True,
                 diagnose=False,
                 retention="3 days",
                 level='INFO',
                 extra: Optional[dict] = None):
    """设置loguru日志记录.

    Args:
        file_name: 日志文件路径及名称
        rotation:
        serialize: 是否序列化, 仅为True时生效
        backtrace:
        diagnose:
        retention:
        level: 日志级别
        extra: 额外信息
    """
    logger.configure(extra=extra)
    setup(file_name,
          rotation=rotation,
          serialize=serialize,
          backtrace=backtrace,
          diagnose=diagnose,
          retention=retention,
          level=level)



from .base import CommonDBConfig
from typing import Optional


class SqliteConfig(CommonDBConfig):
    """Sqlite 配置"""
    driver: Optional[str] = 'sqlite'


class AsyncSqliteConfig(CommonDBConfig):
    """Sqlite async配置"""
    driver: Optional[str] = 'sqlite+aiosqlite'
    is_async: Optional[bool] = True



from typing import Any

from seatools.task import Task, AsyncTask


class TestTask(Task):

    def _run(self, *args, **kwargs) -> Any:
        for i in range(1000):
            self._logger.info('输出: {}', i)

    def _task_name(self) -> str:
        return "测试任务"


class TestAsyncTask(AsyncTask):

    async def _run(self, *args, **kwargs) -> Any:
        for i in range(1000):
            self._logger.info('输出: {}', i)

    def _task_name(self) -> str:
        return "测试async任务"


def test_task():
    task = TestTask()
    task.run()


def test_async_task():
    import asyncio
    task = TestAsyncTask()
    asyncio.run(task.run())



from loguru import logger as log
from seatools.logger import setup as stp, get_loguru_adapter_logging_formatter
import logging

stp('test_logger.log')


@log.catch
def error_fun():
    return 1 / 0


def test_log_catch():
    f = error_fun()
    print(f)


def test_logging():
    filer_handler = logging.FileHandler('demo.log', encoding='utf-8')
    filer_handler.setLevel(logging.DEBUG)
    filer_handler.setFormatter(get_loguru_adapter_logging_formatter()())

    logger = logging.getLogger('a')
    logger.setLevel(logging.DEBUG)
    logger.addHandler(filer_handler)

    logger.info("""啦啦啦
    什么鬼
    \n\n
    """)



import dataclasses
import inspect
import json
import uuid

from seatools import ioc
from seatools.ioc import Autowired, Value
from seatools.ioc.beans.factory import InitializingBean
from seatools.ioc.config import load_config, cfg
from seatools.models import BaseModel


@ioc.Bean  # 默认name会转为小驼峰 randomUuidFunc
def random_uuid_func():
    return uuid.uuid4


@ioc.Bean
def int_bean():
    return 1


@ioc.Bean
class A:
    def __init__(self, uuid_func=ioc.Autowired(value='randomUuidFunc')):
        # 不支持在init中操作uuid_func依赖注入对象
        self._uuid_func = uuid_func
        self._uuid = None

    def __post_construct__(self):
        self._uuid = self._uuid_func()
        print('A.__post_construct__被执行, uuid=', self._uuid)

    def data(self):
        return self._uuid


@ioc.Bean(name='a2', primary=True)
def a():
    return A()


@ioc.Bean
class B(InitializingBean):
    def after_properties_set(self):
        print('B.after_properties_set被执行, a.data() = ', self._a.data())

    def __init__(self, a: A = ioc.Autowired(cls=A)):
        self._a = a

    def data(self):
        return 'B delay {}'.format(self._a.data())


@ioc.Bean
class C:
    a: int
    b: int
    c: str

    def __init__(self, a=ioc.Value('${c.a}'),
                 b=ioc.Value('${c.b}'),
                 c=ioc.Value('${c.c}', default_value=None),
                 ):
        self.a = a
        self.b = b
        self.c = c


class Cp(BaseModel):
    a: int
    b: int
    c: str

def test_ioc():
    # 启动ioc, 扫描当期tests包
    ioc.run(scan_package_names='tests.test_ioc', config_dir='config')
    # context = ioc.get_application_context()
    # v = Autowired('intBean')
    # print(v)
    # a = Autowired('a', cls=A)
    # print(a.data())
    # b = Autowired(cls=B)
    # print(b.data())
    v = Autowired('intBean', cls=int)
    print(v.int == 1)
    c = Autowired(cls=C)
    print(c)
    print(Value('${a}').int)
    print(Value('${b}').int)
    c = Value('${c}', cls=Cp)
    print(c.c)


@ioc.Bean
class AA:

    def hello(self):
        print('hello AA')


@ioc.Bean
class BB:

    def __init__(self, aa: AA):
        self._aa = aa

    def hello(self):
        self._aa.hello()
        print('Hello BB')


def test_depends():
    ioc.run(scan_package_names='tests.test_ioc', config_dir='config')
    b: BB = ioc.Autowired(cls=BB)
    b.hello()



import datetime
import uuid
from abc import ABC, abstractmethod
from typing import Any
from seatools.logger import PrefixLogger


class Task(ABC):
    """任务抽象类封装, 建议任务类业务统一继承该类"""

    def __init__(self):
        self._task_id = self._gen_task_id()
        self._logger = PrefixLogger(self._prefix())

    def run(self, *args, **kwargs) -> Any:
        """任务执行方法"""
        now = datetime.datetime.now()
        self._logger.info('开始处理, 任务参数: args: {}, kwargs: {}', args, kwargs)
        try:
            ans = self._run(*args, **kwargs)
            self._logger.success('处理完成, 本次任务执行耗时: {}, 返回值: {}, 任务参数: args: {}, kwargs: {}, 任务结果: {}',
                            (datetime.datetime.now() - now).total_seconds(), ans, args, kwargs, ans)
            return ans
        except Exception as e:
            self._logger.error('出错, 本次任务执行耗时: {}, 任务参数: args: {}, kwargs: {}, 异常信息: {}',
                               (datetime.datetime.now() - now).total_seconds(), args, kwargs, e)
            raise e

    def _prefix(self):
        return f'{self._task_name()}[{self._task_id}]'

    @abstractmethod
    def _run(self, *args, **kwargs) -> Any:
        """任务逻辑, 子类实现该方法来实现任务逻辑

        Args:
            args: 任务参数
            kwargs: 任务参数
        """
        raise NotImplemented('任务逻辑未实现')

    @abstractmethod
    def _task_name(self) -> str:
        """任务名称, 子类必须实现该方法

        Returns:
            任务名称
        """
        raise NotImplemented('任务名称未实现')

    def _gen_task_id(self):
        return str(uuid.uuid4())


class AsyncTask(ABC):
    """Async任务抽象类封装, 建议任务类业务统一继承该类"""

    def __init__(self):
        self._task_id = self._gen_task_id()
        self._logger = PrefixLogger(self._prefix())

    def _prefix(self):
        return f'{self._task_name()}[{self._task_id}]'

    @abstractmethod
    def _task_name(self) -> str:
        """任务名称, 子类必须实现该方法

        Returns:
            任务名称
        """
        raise NotImplemented('任务名称未实现')

    def _gen_task_id(self):
        return str(uuid.uuid4())

    async def run(self, *args, **kwargs) -> Any:
        """任务执行方法"""
        now = datetime.datetime.now()
        self._logger.info('开始处理, 任务参数: args: {}, kwargs: {}', args, kwargs)
        try:
            ans = await self._run(*args, **kwargs)
            self._logger.success('处理完成, 本次任务执行耗时: {}, 返回值: {}, 任务参数: args: {}, kwargs: {}, 任务结果: {}',
                              (datetime.datetime.now() - now).total_seconds(), ans, args, kwargs, ans)
            return ans
        except Exception as e:
            self._logger.error('出错, 本次任务执行耗时: {}, 任务参数: args: {}, kwargs: {}',
                               (datetime.datetime.now() - now).total_seconds(), args, kwargs)
            raise e

    @abstractmethod
    async def _run(self, *args, **kwargs) -> Any:
        """任务逻辑, 子类实现该方法来实现任务逻辑

        Args:
            args: 任务参数
            kwargs: 任务参数
        """
        raise NotImplemented('任务逻辑未实现')



import datetime
import threading
import time
import traceback
from concurrent.futures import ThreadPoolExecutor
from seatools.mysql import pymysql
from seatools.models import BaseModel
from typing import Optional

pool = ThreadPoolExecutor(200)

SQL_DDL = """
CREATE TABLE `test_pymysql_tbl` (
  `id` bigint(20) unsigned NOT NULL AUTO_INCREMENT COMMENT '主键ID',
  `name` varchar(100) NOT NULL COMMENT '名称',
  `is_delete` tinyint(3) unsigned NOT NULL DEFAULT '0' COMMENT '是否删除',
  `create_time` datetime NOT NULL DEFAULT CURRENT_TIMESTAMP COMMENT '创建时间',
  `update_time` datetime DEFAULT NULL COMMENT '最近一次更新时间',
  PRIMARY KEY (`id`),
  UNIQUE KEY `test_pymysql_tbl_unique` (`name`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COMMENT='测试pymysql表';
"""


class TestPyMysqlModel(BaseModel):
    id: int
    name: str
    is_delete: bool
    create_time: datetime.datetime
    update_time: Optional[datetime.datetime] = None


def save(conn: pymysql.Connection, sql):
    try:
        conn.ping(reconnect=True)
        cursor = conn.cursor()
        cursor.execute(sql)
        conn.commit()
        print('执行sql: ', sql, flush=True)
    except Exception:
        traceback.print_exc()


def test_pymysql():
    conn = pymysql.connect(
        host='localhost',
        port=3306,
        user='root',
        password='123456',
        database='test_db',
    )
    for i in range(10000):
        pool.submit(save, conn, f"""
        insert into test_pymysql_tbl(name) values(test_pymysql_{time.time()})
        """)


def test_pymysql_helper():
    client = pymysql.PyMysqlClient(
        host='localhost',
        port=3306,
        user='root',
        password='123456',
        database='test_db',
    )
    sql = 'select * from test_pymysql_tbl limit 10'
    ans = client.execute(sql, modelclass=TestPyMysqlModel)
    print(ans)
    with client.transaction_client() as t_client:
        ans = t_client.execute(sql, modelclass=TestPyMysqlModel)
        print(ans)



import queue
import threading
from seatools.ioc.beans.factory import BeanFactory
from seatools.ioc.context import ApplicationContext
from seatools.ioc.environment import Environment

_lock = threading.RLock()
# 该队列用来解决在bean factory创建之前就执行注入的内容
_pre_inject_queue = queue.Queue()


class _BeanVars:
    bean_factory: BeanFactory = None
    application_context = None
    environment = None


def new_bean_factory(factory=None):
    with _lock:
        if factory:
            _BeanVars.bean_factory = factory
        else:
            from seatools.ioc.beans.factory import SimpleBeanFactory
            _BeanVars.bean_factory = SimpleBeanFactory()
    _consume_pre_inject_queue()
    return _BeanVars.bean_factory


def get_application_context() -> ApplicationContext:
    if _BeanVars.application_context:
        return _BeanVars.application_context
    with _lock:
        _BeanVars.application_context = ApplicationContext(bean_factory=_BeanVars.bean_factory)
    return _BeanVars.application_context


def get_environment() -> Environment:
    if _BeanVars.environment:
        return _BeanVars.environment
    with _lock:
        context = get_application_context()
        _BeanVars.environment = context.get_bean(name='environment', cls=Environment)
    return _BeanVars.environment


def _get_bean_factory():
    return _BeanVars.bean_factory


def _register_bean(**kwargs):
    """注册bean"""
    # 如果bean工厂已经创建, 则直接注入
    if _BeanVars.bean_factory:
        _BeanVars.bean_factory.register_bean(**kwargs)
        return
    # bean工厂不存在, 则放入队列
    _pre_inject_queue.put(kwargs)


def _consume_pre_inject_queue():
    """消费注入函数队列"""
    while True:
        try:
            kwargs = _pre_inject_queue.get(block=False)
            _BeanVars.bean_factory.register_bean(**kwargs)
        except queue.Empty:
            break



from .base import CommonDBConfig, SqlalchemyConfig
from .sqlite import SqliteConfig, AsyncSqliteConfig
from .mysql import MysqlConfig, AsyncMysqlConfig
from .posgres import PostgresqlConfig, AsyncPostgresqlConfig
from .hive import HiveConfig
from .impyla import ImpylaConfig
from .clickhouse import ClickhouseConfig



from typing import Optional, TypeVar
from seatools.models import BaseModel
import redis
from redis_om import HashModel as _HashModel, JsonModel as _JsonModel, EmbeddedJsonModel as _EmbeddedJsonModel, \
    RedisModel, Field, get_redis_connection

Model = TypeVar("Model", bound="RedisModel")


class HashModel(BaseModel, _HashModel):

    def save(
        self: "Model", pipeline: Optional[redis.client.Pipeline] = None
    ) -> "Model":
        self.check()
        db = self._get_db(pipeline)
        document = self.model_dump(mode='json')

        # filter out values which are `None` because they are not valid in a HSET
        document = {k: v for k, v in document.items() if v is not None}
        # TODO: Wrap any Redis response errors in a custom exception?
        db.hset(self.key(), mapping=document)
        return self


class JsonModel(BaseModel, _JsonModel):
    pass


class EmbeddedJsonModel(BaseModel, _EmbeddedJsonModel):
    pass



from seatools.impala.impyla import ImpylaClient
from loguru import logger
from seatools.models import BaseModel
import datetime
from typing import Optional


class TestModel(BaseModel):
    id: int
    name: str
    desc: str
    create_time: datetime.datetime
    update_time: Optional[datetime.datetime] = None


def test_impyla():
    client = ImpylaClient(host='localhost', port=21050,
                          user='root',
                          database='test_db')
    ans = client.execute("""
SELECT *
FROM test
limit 10;
""")
    logger.info(ans)
    ans = client.execute("""
    SELECT *
    FROM test
    limit 10;
    """, modelclass=TestModel)
    logger.info(ans)



from seatools.notices import SmtpEmailNotice, FeishuRobotNotice, FeishuRobotTextMsg, FeishuRobotPostMsg, \
    FeishuRobotCardMsg, FeishuRobotCardMsgTemplate
from seatools.builders import HtmlBuilder


def test_smtp_email_notice():
    notice = SmtpEmailNotice(smtp_server='smtp.feishu.cn',
                             smtp_port=465,
                             from_email='xxx@xxx.com',
                             from_password='123456',
                             from_name='测试邮箱')
    notice.send_email(title='测试邮件',
                      body=HtmlBuilder(auto_add_html_tag=True
                                       ).h1('测试标题').p('测试').a('测试链接', 'https://www.baidu.com').build(),
                      to_emails=['xxx@xxx.com'],
                      to_cc_mails=['xxx@xxx.com'],
                      attachments=[__file__])


def test_feishu_robot_notice():
    notice = FeishuRobotNotice('xxx',
                               'xxxx')
    # 测试普通消息
    notice.send_msg(FeishuRobotTextMsg(text='测试消息'))
    # 测试富文本消息
    notice.send_msg(FeishuRobotPostMsg(title='测试标题',
                                       content=FeishuRobotPostMsg.content_builder(
                                       ).add_text(
                                           '测试文本'
                                       ).add_a(
                                           '测试链接', 'https://www.baidu.com'
                                       ).add_newline().add_at(
                                           'all'
                                       ).build()))
    # 测试卡片消息
    # 使用卡片消息模板消息生成卡片消息发送(推荐)
    card_template = FeishuRobotCardMsgTemplate(elements=[
        {
            "tag": "div",
            "text": {
                "content": "{{content}}!xxxx",
                "tag": "lark_md"
            }
        }
    ], config={
        "wide_screen_mode": True
    }, header={
        "template": "blue",
        "title": {
            "content": "{{title}}",
            "tag": "plain_text"
        }
    })
    # 使用卡片模板card_template.gen_msg方法生成卡片消息并发送
    notice.send_msg(card_template.gen_msg(elements_map={'content': '测试内容'},
                                          header_map={'title': '测试标题'}))
    # 直接发送卡片消息
    notice.send_msg(FeishuRobotCardMsg(elements=[
        {
            "tag": "div",
            "text": {
                "content": "xxxx",
                "tag": "lark_md"
            }
        }
    ], config={
        "wide_screen_mode": True
    }, header={
        "template": "blue",
        "title": {
            "content": "xxxxx",
            "tag": "plain_text"
        }
    }))



from seatools.models import BaseModel
from clickhouse_driver import connect
from clickhouse_driver.dbapi.cursor import Cursor
from typing import Optional
from typing import Union, Any, List, Tuple, Type


class _DictCursor(Cursor):
    is_query = False

    def execute(self, operation, parameters=None):
        if str(operation).strip().lower().startswith('select'):
            self.is_query = True
        super().execute(operation, parameters=parameters)

    def executemany(self, operation, seq_of_parameters):
        if str(operation).strip().lower().startswith('select'):
            self.is_query = True
        super().executemany(operation, seq_of_parameters)

    def fetchall(self):
        rows = super().fetchall()
        if rows and self.is_query:
            columns = [column.name for column in self.description]
            rows = [dict(zip(columns, row)) for row in rows]
        return rows

    def fetchone(self):
        one = super().fetchone()
        if one and self.is_query:
            columns = [column.name for column in self.description]
            one = dict(zip(columns, one))
        return one

    def fetchmany(self, size=None):
        rows = super().fetchmany(size)
        if rows and self.is_query:
            columns = [column.name for column in self.description]
            rows = [dict(zip(columns, row)) for row in rows]
        return rows


class _ClickhouseConfig(BaseModel):
    host: Optional[str] = 'localhost'
    port: Optional[int] = 9000
    user: Optional[str] = None
    password: Optional[str] = None
    db: Optional[str] = None


class ClickhouseDriverClient:

    def __init__(self,
                 host='localhost',
                 port=9000,
                 user: Optional[str] = None,
                 password: Optional[str] = None,
                 database: Optional[str] = None,
                 **kwargs):
        self._config = _ClickhouseConfig(
            host=host,
            port=port,
            user=user,
            password=password,
            db=database,
        )
        self._kw = kwargs

    def execute(self,
                sql: str,
                args: Union[List[Any], Tuple[Any], None] = None,
                modelclass: Optional[Type[BaseModel]] = None) -> Union[
        List[dict], Tuple[dict], List[BaseModel], Tuple[BaseModel]]:
        conn = connect(
            host=self._config.host,
            port=self._config.port,
            user=self._config.user,
            password=self._config.password,
            database=self._config.db,
            **self._kw
        )
        cursor = conn.cursor(cursor_factory=_DictCursor)
        try:
            cursor.execute(sql, parameters=args)
            # 查询sql转为dict
            # if sql.strip().lower().startswith('select'):
            #     fields = [desc[0] for desc in ans.description]
            ans = cursor.fetchall()
            if ans and modelclass:
                ans = list(map(lambda e: modelclass(**e), ans))
            return ans
        finally:
            if cursor:
                cursor.close()
            if conn:
                conn.close()



# 通用基础实现

__round_mask = 0.00001


def round_(number, decimal: int = 0) -> float:
    """四舍五入, 解决内置round奇进偶舍问题

    Args:
        number: 数值
        decimal: 小数位数

    Returns:
        四舍五入保留的值
    """
    percent = 10**decimal
    number *= percent
    number = round(number + __round_mask / percent)
    # 防止整数转为浮点数
    if decimal > 0:
        number /= percent
    return number



from typing import Optional

from seatools.models import BaseModel


class SqlalchemyConfig(BaseModel):
    """sqlalchemy相关配置"""
    echo: Optional[bool] = False
    # 连接池回收周期
    pool_recycle: Optional[int] = 3600


class CommonDBConfig(BaseModel):
    """通用 DB 配置"""
    name: Optional[str] = None
    host: Optional[str] = None
    port: Optional[int] = None
    user: Optional[str] = None
    password: Optional[str] = None
    db: Optional[str] = None
    # sqlalchemy的schema, 仅使用sqlalchemy需要配置, 例如:sqlite, mysql+pymysql等等
    driver: Optional[str] = None
    # 是否是async连接
    is_async: Optional[bool] = False
    # 是否是ioc primary实例
    primary: Optional[bool] = False
    sqlalchemy: Optional[dict] = None



import os
from seatools.env import get_env


def test_dev_env():
    os.environ['ENV'] = 'dev'
    print(get_env().is_dev())
    print(get_env().is_test())
    print(get_env().is_pro())


def test_test_env():
    os.environ['ENV'] = 'test'
    print(get_env())


def test_pro_env():
    os.environ['ENV'] = 'pro'
    print(get_env())


def test_other_env():
    os.environ['ENV'] = 'xxx'
    print(get_env())



import asyncio

from seatools.retry import Retry, AsyncRetry
from loguru import logger

cnt = 0

GLOBAL_RETRY = Retry(max_attempts=10,  # 最大重试次数
                     delay=1,  # 每次重试的等待间隔, 单位秒
                     log_enable=True,  # 是否打印日志
                     ever_fail_level='WARNING',  # 每次任务失败记录的日志级别(未达到重试上限),
                     fail_level='ERROR',  # 任务重试达到失败上限的日志级别
                     )

GLOBAL_ASYNC_RETRY = AsyncRetry(max_attempts=10,  # 最大重试次数
                                delay=1,  # 每次重试的等待间隔, 单位秒
                                log_enable=True,  # 是否打印日志
                                ever_fail_level='WARNING',  # 每次任务失败记录的日志级别(未达到重试上限),
                                fail_level='ERROR',  # 任务重试达到失败上限的日志级别
                                )


def fail3(a: int, b: int):
    logger.info('请求参数a: {}, b: {}', a, b)
    global cnt
    if cnt < 3:
        cnt += 1
        raise Exception('异常')
    logger.info('cnt: {}', cnt)
    return cnt


def test_retry():
    _retry = Retry(max_attempts=10,  # 最大重试次数
                   delay=1,  # 每次重试的等待间隔, 单位秒
                   log_enable=True,  # 是否打印日志
                   ever_fail_level='WARNING',  # 每次任务失败记录的日志级别(未达到重试上限),
                   fail_level='ERROR',  # 任务重试达到失败上限的日志级别
                   )
    logger.info(_retry.do(fail3, 1, b=5))


@GLOBAL_RETRY.retry
def fail3_wrapper(a: int, b: int):
    logger.info('请求参数a: {}, b: {}', a, b)
    global cnt
    if cnt < 3:
        cnt += 1
        raise Exception('异常')
    logger.info('cnt: {}', cnt)
    return cnt


def test_retry2():
    fail3_wrapper(1, 2)


async def test(a: int):
    raise Exception(f'异常{a}')


def test_async_retry():
    _async_retry = AsyncRetry(max_attempts=3,  # 最大重试次数
                              delay=1,  # 每次重试的等待间隔, 单位秒
                              log_enable=True,  # 是否打印日志
                              ever_fail_level='WARNING',  # 每次任务失败记录的日志级别(未达到重试上限),
                              fail_level='ERROR',  # 任务重试达到失败上限的日志级别
                              )
    asyncio.run(_async_retry.do(test, 3))


@GLOBAL_ASYNC_RETRY.retry
async def fail3_async_wrapper(a: int, b: int):
    logger.info('请求参数a: {}, b: {}', a, b)
    global cnt
    if cnt < 3:
        cnt += 1
        raise Exception('异常')
    logger.info('cnt: {}', cnt)
    return cnt


def test_async_retry2():
    asyncio.run(fail3_async_wrapper(1, 2))



import re


def to_camel_case(name: str, upper_case: bool = False) -> str:
    """转为驼峰

    Args:
        name: 转换的名称
        upper_case: True=大驼峰, False=小驼峰
    """
    if not name:
        return name
    if len(name) == 1:
        return name.upper() if upper_case else name.lower()
    items = re.split(r'[_\s]+', name)
    if len(items) == 1:
        return _to_camel_case_item(items[0], upper_case=upper_case)
    return _to_camel_case_item(items[0], upper_case=upper_case) + ''.join(_to_camel_case_item(x, upper_case=True) for x
                                                                          in items[1:])


def _to_camel_case_item(name: str, upper_case: bool = False) -> str:
    if not name:
        return name
    if len(name) == 1:
        return name.upper() if upper_case else name.lower()
    return name.title() if upper_case else (name[0].lower() + name[1:])



from loguru import logger as log
from typing import Union, List

from .base import DataFileLoader, DataFileExtractor
import xml.etree.ElementTree as ET


class XmlDataFileLoader(DataFileLoader):

    def _convert_config_data_to_python(self,
                                       config_data: Union[str, bytes],
                                       encoding: str = 'utf-8',
                                       **kwargs) -> Union[dict, tuple, list]:
        try:
            import xmltodict
            return xmltodict.parse(config_data)
        except ImportError as e:
            log.warning('未安装xmltodict模块, 无法解析xml格式数据')
            raise e


class XmlDataFileExtractor(DataFileExtractor):

    def _convert_data_to_str(self, data: Union[dict, List[dict]],
                             encoding='utf-8', **kwargs):
        def dict_to_xml(tag, d):
            """将字典转换为XML元素"""
            elem = ET.Element(tag)
            for key, val in d.items():
                if isinstance(val, dict):
                    # 如果值是字典，递归创建子元素
                    child = dict_to_xml(key, val)
                elif isinstance(val, list):
                    # 如果值是列表，遍历列表并为每个元素创建子元素
                    for item in val:
                        if isinstance(item, dict):
                            child = dict_to_xml(key, item)
                        else:
                            child = ET.Element(key)
                            child.text = self._extract_value(item)
                        elem.append(child)
                    continue
                else:
                    # 如果值不是字典或列表，创建文本元素
                    child = ET.Element(key)
                    child.text = self._extract_value(val)
                elem.append(child)
            return elem

        def dict_to_xml_string(data_dict):
            """将整个字典转换为XML字符串"""
            # 创建字典中所有顶级键的元素
            elements = []
            for key, value in data_dict.items():
                child = dict_to_xml(key, value)
                elements.append(child)
            return '\n'.join([ET.tostring(element, encoding=encoding).decode(encoding) for element in elements])
        if isinstance(data, dict):
            data = [data]
        ans = ''
        for item in data:
            ans += '{}\n'.format(dict_to_xml_string(item))
        return ans




from typing import Union, Tuple, Type, TypeVar

_T = TypeVar('_T', bound=Exception)


def do_uncaught_exceptions(func, *args,
                           uncaught_exceptions: Union[
                               Type[_T], Tuple[Type[_T]]] = Exception,
                           **kwargs):
    """执行并忽略所有指定的异常

    Args:
        func: 执行的方法/函数
        uncaught_exceptions: 执行方法不处理的异常

    Returns:
        方法/函数执行成功返回对应的返回值, 失败返回None
    """
    try:
        return func(*args, **kwargs)
    except uncaught_exceptions:
        return None



from .base import CommonDBConfig
from typing import Optional


class ImpylaConfig(CommonDBConfig):
    """Impala 配置"""
    host: Optional[str] = '127.0.0.1'
    port: Optional[int] = 21050
    driver: Optional[str] = 'impala'



from seatools.algorithm.similarity import cos_similarity
from loguru import logger


def test_cos_similarity():
    a = [0, 0, 0, 1, 1, 0, 1, 0, 1]
    b = [0, 1, 0, 0, 0, 0, 0, 0, 0]
    c = [0, 0, 1, 0, 0, 0, 0, 0, 0]
    d = [1, 1, 1, 1, 1, 1, 1, 0, 1]
    logger.success('')
    logger.success(cos_similarity(a, d))
    logger.success(cos_similarity(b, d))
    logger.success(cos_similarity(c, d))




import inspect

from seatools.utils import list_utils
from seatools.ioc.base import _register_bean
from seatools.ioc.utils.name_utils import to_camel_case


def bean(*args, name: str = None, primary=False):
    """bean装饰器

    Args:
        name: bean name
        primary: 是否默认, 当通过类型获取bean有多个bean时， 若无有primary或存在多个primary的bean将抛出异常, 仅当有一个primary时正常返回
    """

    def wrapper(fc=None):
        # 注册bean
        _register_bean(name=name, cls=fc, primary=primary)
        fc.__bean__ = True
        return fc

    if len(args) == 1 and (inspect.isclass(args[0]) or inspect.isfunction(args[0])):
        return wrapper(args[0])

    name = list_utils.get(args, 0, name)
    primary = list_utils.get(args, 1, primary)
    return wrapper


Bean = bean


def configuration_properties_bean(*args, prop: str = None, name: str = None, primary=False):
    """配置属性bean. 仅支持pydantic model, dataclass, 同配置的构造函数, 仅支持非必填构造器, 若需要注入的构造参数默认值请使用unique_tools.ioc.Autowired()代替.

    Args:
        prop: 属性描述
        name: bean name
        primary: 是否默认, 当通过类型获取bean有多个bean时， 若无有primary或存在多个primary的bean将抛出异常, 仅当有一个primary时正常返回
    """

    def wrapper(fc=None):
        # 注册bean
        if inspect.isfunction(fc):
            fc = type(fc())
        elif not inspect.isclass(fc):
            fc = type(fc)

        def lazy_register_bean():
            from seatools.ioc import get_environment
            from seatools.ioc.environment import Environment
            e: Environment = get_environment()
            return e.get_property(prop, fc)

        _register_bean(name=name or to_camel_case(fc.__name__, False), cls=lazy_register_bean, primary=primary)
        fc.__bean__ = True
        return fc

    if len(args) == 1 and (inspect.isclass(args[0]) or inspect.isfunction(args[0])):
        return wrapper(args[0])

    prop = list_utils.get(args, 0, prop)
    name = list_utils.get(args, 1, name)
    primary = list_utils.get(args, 2, primary)
    return wrapper


ConfigurationPropertiesBean = configuration_properties_bean



from seatools.builders import HtmlBuilder, MarkdownBuilder


def test_html_builder():
    b = HtmlBuilder(auto_add_html_tag=True)
    b.h1('标题').p('啦啦啦').no_tag('hello').div(HtmlBuilder().p('11111'))
    html = b.build()
    print(html)


def test_markdown_builder():
    m = MarkdownBuilder()
    m.h1('标题').p('啦啦啦').no_tag('hello').a('测试链接', 'https://www.baidu.com')
    markdown = m.build()
    print(markdown)



from sqlalchemy import BigInteger, String, DateTime, select, update, delete
from sqlalchemy.orm import Mapped, mapped_column
from loguru import logger
from seatools.sqlalchemy import SqlAlchemyClient, Base, AsyncSqlAlchemyClient
import datetime
from seatools.models import BaseModel


class User(Base):
    __tablename__ = 'user'

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)
    name: Mapped[str] = mapped_column(String(30), unique=True, nullable=False)
    create_time: Mapped[datetime.datetime] = mapped_column(DateTime)


class UserModel(BaseModel):
    id: int
    name: str
    create_time:datetime.datetime
    lalala: str


def get_helper():
    return SqlAlchemyClient('sqlite:///example.db', echo=True)


def test_create_model():
    Base.metadata.create_all(bind=get_helper().engine())


def test_insert():
    u = UserModel(id=2, name='测试用户2', create_time=datetime.datetime.now(), lalala='hee')
    helper = get_helper()
    with helper.session() as session:
        session.add(User.from_model(u))


def test_query():
    helper = get_helper()
    with helper.session() as session:
        users = session.execute(select(User).where(User.id >= 1)).scalars().all()
        users = [user.to_model(UserModel) for user in users]
        logger.info(users)


def test_update():
    helper = get_helper()
    with helper.session() as session:
        session.execute(update(User).where(User.id == 1).values(name='测试用户啦啦啦'))


def test_delete():
    helper = get_helper()
    with helper.session() as session:
        session.execute(delete(User).where(User.id == 2))


async def test_async_engine():
    helper = AsyncSqlAlchemyClient('sqlite+aiosqlite:///example.db', echo=True)
    async with helper.session() as session:
        session.add(User(id=1, name='测试用户1', create_time=datetime.datetime.now()))
        users = (await session.execute(select(User).where(User.id >= 1))).scalars().all()
        logger.info(users)



from seatools.models import BaseModel, R, PageR, PageModel
from loguru import logger
from pydantic import BaseModel as PydanticBaseModel
import datetime


def test_models():
    class AModel(BaseModel):
        a: int
        b: datetime.datetime
        c: datetime.date
        d: datetime.time

    a = AModel(a=1, b='2023-12-31 00:23:21', c='2023-12-31', d='00:23:31')
    print(a.model_dump_json())


def test_r_models():
    print(PageR.ok(data=PageModel()))
    print(R.ok('hello'))


def test_model_to_str():
    class Old(PydanticBaseModel):
        a: str

    class New(BaseModel):
        a: str

    try:
        o = Old(a=1)
        logger.info(o)
    except Exception as e:
        logger.error('pydantic model 异常')

    try:
        n = New(a=True)
        logger.info(n)
    except Exception as e:
        logger.error('自定义 model 异常')
